from enum import Enum
from typing import Any

import xmlschema
from lxml import etree

from to_cei import config


class Schema(str, Enum):
    CEI = config.CEI_NS


class Validator:
    def __validate(self, element: etree._Element, schema: Schema) -> None:
        xsd_content = config.file_cache.get(schema.value)
        if xsd_content:
            xsd = xmlschema.XMLSchema11(xsd_content)
            resource: Any = element
            xsd.validate(resource)

    def validate_cei(self, element: etree._Element):
        return self.__validate(element, Schema.CEI)



import pathlib
from datetime import datetime
from typing import List

import pytest
from astropy.time.core import Time
from lxml import etree

from pytest_helpers import xp, xps
from to_cei.charter import NO_DATE_TEXT, NO_DATE_VALUE, Charter
from to_cei.config import (CEI, CHARTER_NSS, SCHEMA_LOCATION,
                           SCHEMA_LOCATION_QNAME)
from to_cei.helpers import ln
from to_cei.seal import Seal
from to_cei.validator import Validator

# --------------------------------------------------------------------#
#                         Charter as a whole                         #
# --------------------------------------------------------------------#


def test_has_correct_base_structure():
    direct_children = xp(Charter(id_text="1"), "/cei:text/child::*")
    assert len(direct_children) == 3
    assert ln(direct_children[0]) == "front"
    assert ln(direct_children[1]) == "body"
    assert ln(direct_children[2]) == "back"


def test_is_valid_charter():
    charter = Charter(
        "1307 II 22",
        abstract="Konrad von Lintz, Caplan zu St. Pankraz, beurkundet den vorstehenden Vertrag mit Heinrich, des Praitenvelders Schreiber.",
        abstract_sources=["HAUSWIRTH, Schotten (=FRA II/18, 1859) S. 123, Nr. 103"],
        archive="Stiftsarchiv Schotten, Wien (http://www.schottenstift.at)",
        chancellary_remarks=[
            "commissio domini imperatoris in consilio",
            "Jüngerer Dorsualvermerk mit Regest",
        ],
        comments="The diplomatic analysis is inconclusive",
        condition="Beschädigtes Pergament",
        date="1307 II 22",
        date_quote="an sand peters tage in der vasten, als er avf den stvl ze Rome gesatz wart",
        date_value=Time("1307-02-22", format="isot", scale="ut1"),
        dimensions="20x20cm",
        external_link="https://example.com/charters/1",
        footnotes=["Siehe RI #1234", "Abweichend von Nr. 15"],
        graphic_urls=["K.._MOM-Bilddateien._~Schottenjpgweb._~StAS__13070222-2.jpg"],
        index=["Arenga", CEI.index("Insulare Minuskel")],
        index_geo_features=["Leithagebirge", CEI.geogName("Donau")],
        index_organizations=["Bistum Passau", CEI.orgName("Erzbistum Salzburg")],
        index_persons=[
            "Hubert, der Schuster",
            CEI.persName("Antonia Müllerin, des Müllers Frau"),
        ],
        index_places=["Wien", CEI.placeName("Wiener Neustadt")],
        issued_place="Wiener Neustadt",
        issuer="Konrad von Lintz",
        language="Deutsch",
        material="Pergament",
        literature="HAUSWIRTH, Schotten (=FRA II/18, 1859) S. 123, Nr. 103",
        literature_abstracts="[RI XIII] H. 4 n. 778, in: Regesta Imperii Online, URI: http://www.regesta-imperii.de/id/1477-04-02_1_0_13_4_0_10350_778 (Abgerufen am 13.11.2016)",
        literature_depictions="ADEVA-Verlag, Die Ostarrichi-Urkunde – Luxus-Ausgabe. Ausstattung und Preise, In: ADEVA Home (2012), online unter <http://www.adeva.com/faks_detail_bibl.asp?id=127> (12.12.2012)",
        literature_editions=[
            "Urkunden der burgundischen Rudolfinger, ed. Theodor Schieffer, Monumenta Germaniae Historica, Diplomata, 2A, Regum Burgundiae e stirpe Rudolfina diplomata et acta, München 1977; in der Folge: MGH DD Burg. 103."
        ],
        literature_secondary=[
            "HAUSWIRTH, Schotten (=FRA II/18, 1859) S. 123-124",
            "HAUSWIRTH, Schotten (=FRA II/18, 1859) S. 127, Nr. 105",
        ],
        notarial_authentication="Albertus Magnus",
        seals="2 Siegel",
        recipient="Heinrich, des Praitenvelders Schreiber",
        tradition="orig.",
        transcription="Ich Hainrich, des Praitenvelder Schreiber, [...] ze Rome gesatz wart.",
        transcription_sources="HAUSWIRTH, Schotten (=FRA II/18, 1859) S. 123-124",
        witnesses=["Franz von Ehrlingen", CEI.persName("Ulrich der Schneider")],
    )
    Validator().validate_cei(charter.to_xml())


def test_writes_correct_file(tmp_path):
    d = tmp_path
    charter = Charter("1A")
    charter.to_file(d)
    out = pathlib.Path(d, "1A.cei.xml")
    assert out.is_file()
    written = etree.parse(str(out))
    Validator().validate_cei(written.getroot())


def test_add_schema_location_is_respected():
    charter = Charter("1A")
    assert (
        charter.to_xml(add_schema_location=True).get(SCHEMA_LOCATION_QNAME)
        == SCHEMA_LOCATION
    )
    assert charter.to_xml(add_schema_location=False).get(SCHEMA_LOCATION_QNAME) == None


# --------------------------------------------------------------------#
#                          Charter abstract                          #
# --------------------------------------------------------------------#


def test_has_correct_text_abstract():
    abstract = (
        "Konrad von Lintz, Caplan zu St. Pankraz, beurkundet den vorstehenden Vertrag."
    )
    charter = Charter(id_text="1", abstract=abstract)
    assert charter.abstract == abstract
    abstract_xml = xps(charter, "/cei:text/cei:body/cei:chDesc/cei:abstract")
    assert abstract_xml.text == abstract


def test_has_correct_xml_abstract():
    pers_name = "Konrad von Lintz"
    abstract = CEI.abstract(
        CEI.persName(pers_name),
        ", Caplan zu St. Pankraz, beurkundet den vorstehenden Vertrag.",
    )
    charter = Charter(id_text="1", abstract=abstract)
    assert charter.abstract == abstract
    abstract_xml = xps(charter, "/cei:text/cei:body/cei:chDesc/cei:abstract")
    assert abstract_xml.text == abstract.text
    pers_name_xml = xps(
        charter, "/cei:text/cei:body/cei:chDesc/cei:abstract/cei:persName"
    )
    assert pers_name_xml.text == pers_name


def test_has_no_abstract_for_empty_text():
    abstract = ""
    charter = Charter(id_text="1", abstract=abstract)
    assert charter.abstract == None


def test_raises_exception_for_incorrect_xml_abstract():
    incorrect_element = CEI.persName("A person")
    with pytest.raises(ValueError):
        Charter(id_text="1", abstract=incorrect_element)


# --------------------------------------------------------------------#
#                          Charter archive                           #
# --------------------------------------------------------------------#


def test_has_correct_charter_archive():
    archive = "Archive 1"
    charter = Charter(id_text="1", archive=archive)
    assert charter.archive == archive
    archive_xml = xps(
        charter,
        "/cei:text/cei:body/cei:chDesc/cei:witnessOrig/cei:archIdentifier/cei:arch",
    )
    assert archive_xml.text == archive


def test_has_no_archive_for_empty_text():
    archive = ""
    charter = Charter(id_text="1", archive=archive)
    assert charter.archive == None


# --------------------------------------------------------------------#
#                       Charter bibliographies                       #
# --------------------------------------------------------------------#


def test_has_correct_abstract_bibl():
    bibl_text = "Bibl a"
    charter = Charter(
        id_text="1",
        abstract_sources=bibl_text,
    )
    assert isinstance(charter.abstract_sources, List)
    bibl = xps(charter, "/cei:text/cei:front/cei:sourceDesc/cei:sourceDescRegest/*")
    assert bibl.text == bibl_text


def test_has_no_sources_for_empty_string():
    bibl_texts = ""
    sources = xp(
        Charter(
            id_text="1",
            abstract_sources=bibl_texts,
        ),
        "/cei:text/cei:front/cei:sourceDesc/cei:sourceDescRegest/*",
    )
    assert len(sources) == 0


def test_has_correct_abstract_sources():
    bibl_texts = ["Bibl a", "Bibl b"]
    sources = xp(
        Charter(
            id_text="1",
            abstract_sources=bibl_texts,
        ),
        "/cei:text/cei:front/cei:sourceDesc/cei:sourceDescRegest/*",
    )
    assert len(sources) == 2
    assert sources[0].text == bibl_texts[0]
    assert sources[1].text == bibl_texts[1]


def test_has_correct_transcription_source():
    bibl_text = "Bibl a"
    charter = Charter(
        id_text="1",
        transcription_sources=bibl_text,
    )
    assert isinstance(charter.transcription_sources, List)
    sources = xps(
        charter, "/cei:text/cei:front/cei:sourceDesc/cei:sourceDescVolltext/*"
    )
    assert sources.text == bibl_text


def test_has_no_transcription_sources_for_empty_text():
    bibl_text = ""
    charter = Charter(
        id_text="1",
        transcription_sources=bibl_text,
    )
    assert len(charter.transcription_sources) == 0


def test_has_correct_transcription_sources():
    bibl_texts = ["Bibl a", "Bibl b"]
    sources = xp(
        Charter(
            id_text="1",
            transcription_sources=bibl_texts,
        ),
        "/cei:text/cei:front/cei:sourceDesc/cei:sourceDescVolltext/*",
    )
    assert len(sources) == 2
    assert sources[0].text == bibl_texts[0]
    assert sources[1].text == bibl_texts[1]


# --------------------------------------------------------------------#
#                    Charter chancellary remarks                     #
# --------------------------------------------------------------------#


def test_has_correct_single_chancellary_remark():
    chancellary_remarks = "Remark"
    charter = Charter(id_text="1", chancellary_remarks=chancellary_remarks)
    assert isinstance(charter.chancellary_remarks, List)
    assert charter.chancellary_remarks[0] == chancellary_remarks
    nota = xps(charter, "/cei:text/cei:body/cei:chDesc/cei:witnessOrig/cei:nota")
    assert nota.text == chancellary_remarks


def test_has_correct_chancellary_remarks_list():
    chancellary_remarks = ["Remark a", "Remark b"]
    charter = Charter(id_text="1", chancellary_remarks=chancellary_remarks)
    assert charter.chancellary_remarks == chancellary_remarks
    nota = xp(charter, "/cei:text/cei:body/cei:chDesc/cei:witnessOrig/cei:nota")
    assert len(nota) == 2
    assert nota[0].text == chancellary_remarks[0]
    assert nota[1].text == chancellary_remarks[1]


def test_has_no_chancellary_remarks_for_empty_text():
    chancellary_remarks = ""
    charter = Charter(id_text="1", chancellary_remarks=chancellary_remarks)
    assert len(charter.chancellary_remarks) == 0


# --------------------------------------------------------------------#
#                          Charter comments                          #
# --------------------------------------------------------------------#


def test_has_correct_comments():
    comments = ["Comment a", "Comment b"]
    charter = Charter(id_text="1", comments=comments)
    assert charter.comments == comments
    paragraphs = xp(
        charter,
        "/cei:text/cei:body/cei:chDesc/cei:diplomaticAnalysis/cei:p",
    )
    assert len(paragraphs) == 2
    assert paragraphs[0].text == comments[0]
    assert paragraphs[1].text == comments[1]


def test_has_no_comments_for_empty_string():
    comments = ""
    charter = Charter(id_text="1", comments=comments)
    assert len(charter.comments) == 0


# --------------------------------------------------------------------#
#                         Charter condition                          #
# --------------------------------------------------------------------#


def test_has_correct_charter_condition():
    condition = "Charter condition"
    charter = Charter(id_text="1", condition=condition)
    assert charter.condition == condition
    condition_xml = xps(
        charter,
        "/cei:text/cei:body/cei:chDesc/cei:witnessOrig/cei:physicalDesc/cei:condition",
    )
    assert condition_xml.text == condition


def test_has_no_condition_for_empty_text():
    condition = ""
    charter = Charter(id_text="1", condition=condition)
    assert charter.condition == None


# --------------------------------------------------------------------#
#                            Charter date                             #
# --------------------------------------------------------------------#


def test_has_correct_date_range_with_99999999():
    charter = Charter(id_text="1", date="unknown", date_value=("99999999", "99999999"))
    assert charter.date == "unknown"
    assert charter.date_value == None
    date_xml = xps(charter, "/cei:text/cei:body/cei:chDesc/cei:issued/cei:date")
    assert date_xml.text == "unknown"
    assert date_xml.get("value") == NO_DATE_VALUE


def test_has_no_date_for_empty_text():
    charter = Charter(id_text="1", date="")
    assert charter.date == None
    assert charter.date_value == None


def test_has_no_date_for_empty_value():
    charter = Charter(id_text="1", date="unknown", date_value="")
    assert charter.date == "unknown"
    assert charter.date_value == None
    date_xml = xps(charter, "/cei:text/cei:body/cei:chDesc/cei:issued/cei:date")
    assert date_xml.text == "unknown"
    assert date_xml.get("value") == NO_DATE_VALUE


def test_has_correct_date_with_99999999():
    charter = Charter(id_text="1", date="unknown", date_value="99999999")
    assert charter.date == "unknown"
    assert charter.date_value == None
    date_xml = xps(charter, "/cei:text/cei:body/cei:chDesc/cei:issued/cei:date")
    assert date_xml.text == "unknown"
    assert date_xml.get("value") == NO_DATE_VALUE


def test_has_correct_date_range_with_datetime():
    charter = Charter(
        id_text="1",
        date="1300",
        date_value=(datetime(1300, 1, 1), datetime(1300, 12, 31)),
    )
    assert charter.date == "1300"
    assert charter.date_value == (
        Time("1300-01-01", format="isot", scale="ut1"),
        Time("1300-12-31", format="isot", scale="ut1"),
    )
    date_xml = xps(charter, "/cei:text/cei:body/cei:chDesc/cei:issued/cei:dateRange")
    assert date_xml.text == "1300"
    assert date_xml.get("from") == "13000101"
    assert date_xml.get("to") == "13001231"


def test_has_correct_date_with_datetime():
    charter = Charter(
        id_text="1",
        date="1. 1. 1300",
        date_value=datetime(1300, 1, 1),
    )
    assert charter.date == "1. 1. 1300"
    assert charter.date_value == Time("1300-01-01", format="isot", scale="ut1")
    date_xml = xps(charter, "/cei:text/cei:body/cei:chDesc/cei:issued/cei:date")
    assert date_xml.text == "1. 1. 1300"
    assert date_xml.get("value") == "13000101"


def test_has_correct_date_range_with_iso():
    charter = Charter(id_text="1", date="1300", date_value=("1300-01-01", "1300-12-31"))
    assert charter.date == "1300"
    assert charter.date_value == (
        Time("1300-01-01", format="isot", scale="ut1"),
        Time("1300-12-31", format="isot", scale="ut1"),
    )
    date_xml = xps(charter, "/cei:text/cei:body/cei:chDesc/cei:issued/cei:dateRange")
    assert date_xml.text == "1300"
    assert date_xml.get("from") == "13000101"
    assert date_xml.get("to") == "13001231"


def test_has_correct_date_with_iso():
    charter = Charter(id_text="1", date="12. 1. 1342", date_value="1342-01-12")
    assert charter.date == "12. 1. 1342"
    assert charter.date_value == Time("1342-01-12", format="isot", scale="ut1")
    date_xml = xps(charter, "/cei:text/cei:body/cei:chDesc/cei:issued/cei:date")
    assert date_xml.text == "12. 1. 1342"
    assert date_xml.get("value") == "13420112"


def test_has_correct_date_range_with_Time():
    charter = Charter(
        id_text="1",
        date="1300",
        date_value=(
            Time({"year": 1300, "month": 1, "day": 1}, scale="ut1"),
            Time({"year": 1300, "month": 12, "day": 31}, scale="ut1"),
        ),
    )
    assert charter.date == "1300"
    assert charter.date_value == (
        Time("1300-01-01", format="isot", scale="ut1"),
        Time("1300-12-31", format="isot", scale="ut1"),
    )
    date_xml = xps(charter, "/cei:text/cei:body/cei:chDesc/cei:issued/cei:dateRange")
    assert date_xml.text == "1300"
    assert date_xml.get("from") == "13000101"
    assert date_xml.get("to") == "13001231"


def test_has_correct_date_with_Time():
    charter = Charter(
        id_text="1",
        date="1. 1. 1300",
        date_value=Time({"year": 1300, "month": 1, "day": 1}, scale="ut1"),
    )
    assert charter.date == "1. 1. 1300"
    assert charter.date_value == Time("1300-01-01", format="isot", scale="ut1")
    date_xml = xps(charter, "/cei:text/cei:body/cei:chDesc/cei:issued/cei:date")
    assert date_xml.text == "1. 1. 1300"
    assert date_xml.get("value") == "13000101"


def test_has_correct_empty_date():
    charter = Charter(id_text="1")
    assert charter.date == None
    assert charter.date_value == None
    date_xml = xps(charter, "/cei:text/cei:body/cei:chDesc/cei:issued/cei:date")
    assert date_xml.text == NO_DATE_TEXT
    assert date_xml.get("value") == NO_DATE_VALUE


def test_has_correct_empty_date_range_text():
    charter = Charter(id_text="1", date_value=("1300-01-01", "1300-12-31"))
    assert charter.date == None
    assert charter.date_value == (
        Time("1300-01-01", format="isot", scale="ut1"),
        Time("1300-12-31", format="isot", scale="ut1"),
    )
    date_xml = xps(charter, "/cei:text/cei:body/cei:chDesc/cei:issued/cei:dateRange")
    assert date_xml.text == "+01300-01-01 - +01300-12-31"
    assert date_xml.get("from") == "13000101"
    assert date_xml.get("to") == "13001231"


def test_has_correct_empty_date_text():
    charter = Charter(id_text="1", date_value="1342-01-12")
    assert charter.date == None
    assert charter.date_value == Time("1342-01-12", format="isot", scale="ut1")
    date_xml = xps(charter, "/cei:text/cei:body/cei:chDesc/cei:issued/cei:date")
    assert date_xml.text == "+01342-01-12"
    assert date_xml.get("value") == "13420112"


def test_has_correct_empty_date_value():
    text = "Sine dato"
    charter = Charter(id_text="1", date=text)
    assert charter.date == text
    assert charter.date_value == None
    date_xml = xps(charter, "/cei:text/cei:body/cei:chDesc/cei:issued/cei:date")
    assert date_xml.text == text
    assert date_xml.get("value") == NO_DATE_VALUE


def test_has_correct_xml_date():
    date_text = "12. 10. 1789"
    date_value = "17980101"
    date = CEI.date(date_text, {"value": date_value})
    charter = Charter(id_text="1", date=date)
    assert charter.date == date
    date_xml = xps(charter, "/cei:text/cei:body/cei:chDesc/cei:issued/cei:date")
    assert date_xml.text == date_text
    assert date_xml.get("value") == date_value


def test_has_correct_xml_date_range():
    date_text = "some time in 1789"
    date_from = "17980101"
    date_to = "17981231"
    date = CEI.dateRange(date_text, {"from": date_from, "to": date_to})
    charter = Charter(id_text="1", date=date)
    assert charter.date == date
    date_xml = xps(charter, "/cei:text/cei:body/cei:chDesc/cei:issued/cei:dateRange")
    assert date_xml.text == date_text
    assert date_xml.get("from") == date_from
    assert date_xml.get("to") == date_to


def test_raises_exception_when_initializing_with_xml_date_and_value():
    date_value = "17980101"
    date = CEI.date("12. 10. 1789", {"value": date_value})
    with pytest.raises(ValueError):
        Charter(id_text="1", date=date, date_value=date_value)


def test_raises_exception_for_invalid_date_value():
    with pytest.raises(ValueError):
        Charter(
            id_text="1",
            date="in 1789",
            date_value="17980231",  # 31st February doesn't exist
        )


def test_raises_exception_for_invalid_date_value_in_iso_format():
    with pytest.raises(ValueError):
        Charter(
            id_text="1",
            date="in 1789",
            date_value="1798-02-311",  # One additional digit
        )


def test_raises_exception_for_invalid_date_value_in_mom_format():
    with pytest.raises(ValueError):
        Charter(
            id_text="1",
            date="in 1789",
            date_value="179802311",  # One additional digit
        )


def test_raises_exception_for_incorrect_date_value_pair():
    with pytest.raises(ValueError):
        Charter(
            id_text="1",
            date="in 1789",
            date_value=("17980101", datetime(1789, 12, 31)),
        )


def test_raises_exception_for_incorrect_xml_date():
    incorrect_element = CEI.persName("A person")
    with pytest.raises(ValueError):
        Charter(id_text="1", date=incorrect_element)


def test_raises_exception_when_setting_date_value_for_xml_date():
    date_from = "17980101"
    date_to = "17981231"
    date = CEI.dateRange("12. 10. 1789", {"from": date_from, "to": date_to})
    charter = Charter(id_text="1", date=date)
    with pytest.raises(ValueError):
        charter.date_value = (date_from, date_to)


# --------------------------------------------------------------------#
#                         Charter date quote                         #
# --------------------------------------------------------------------#


def test_has_correct_text_date_quote():
    date_quote = "A date quote"
    charter = Charter(id_text="1", date_quote=date_quote)
    assert charter.date_quote == date_quote
    date_quote_xml = xps(
        charter,
        "/cei:text/cei:body/cei:chDesc/cei:diplomaticAnalysis/cei:quoteOriginaldatierung",
    )
    assert date_quote_xml.text == date_quote


def test_has_no_quote_for_empty_text():
    date_quote = ""
    charter = Charter(id_text="1", date_quote=date_quote)
    assert charter.date_quote == None


def test_has_correct_xml_date_quote():
    date_quote = CEI.quoteOriginaldatierung(
        "Original dating with ", CEI.sup("a"), " superscript"
    )
    charter = Charter(id_text="1", date_quote=date_quote)
    assert charter.date_quote == date_quote
    superscript_xml = xps(
        charter,
        "/cei:text/cei:body/cei:chDesc/cei:diplomaticAnalysis/cei:quoteOriginaldatierung/cei:sup",
    )
    assert superscript_xml.text == "a"


def test_raises_exception_for_incorrect_xml_date_quote():
    incorrect_element = CEI.persName("A person")
    with pytest.raises(ValueError):
        Charter(id_text="1", date_quote=incorrect_element)


# --------------------------------------------------------------------#
#                         Charter dimensions                         #
# --------------------------------------------------------------------#


def test_has_correct_charter_dimensions():
    dimensions = "Charter dimensions"
    charter = Charter(id_text="1", dimensions=dimensions)
    assert charter.dimensions == dimensions
    dimensions_xml = xps(
        charter,
        "/cei:text/cei:body/cei:chDesc/cei:witnessOrig/cei:physicalDesc/cei:dimensions",
    )
    assert dimensions_xml.text == dimensions


def test_has_no_dimensions_for_empty_text():
    dimensions = ""
    charter = Charter(id_text="1", dimensions=dimensions)
    assert charter.dimensions == None


# --------------------------------------------------------------------#
#                        Charter external url                        #
# --------------------------------------------------------------------#


def test_has_correct_external_url():
    external_link = "https://example.com/charters/1"
    charter = Charter(id_text="1", external_link=external_link)
    assert charter.external_link == external_link
    external_link_xml = xps(
        charter,
        "/cei:text/cei:body/cei:chDesc/cei:witnessOrig/cei:archIdentifier/cei:ref",
    )
    assert external_link_xml.get("target") == external_link


def test_has_no_external_url_for_empty_text():
    external_link = ""
    charter = Charter(id_text="1", external_link=external_link)
    assert charter.external_link == None


def test_raises_exception_for_invalid_external_link():
    localhost = "http://localhost"
    with pytest.raises(ValueError):
        Charter(id_text="1", external_link=localhost)


# --------------------------------------------------------------------#
#                     Charter figures / graphics                     #
# --------------------------------------------------------------------#


def test_has_correct_list_graphic_urls():
    graphic_urls = ["Figure 1.jgp", "figure_2.png"]
    charter = Charter(id_text="1", graphic_urls=graphic_urls)
    assert charter.graphic_urls == graphic_urls
    graphics_xml = xp(
        charter, "/cei:text/cei:body/cei:chDesc/cei:witnessOrig/cei:figure/cei:graphic"
    )
    assert len(graphics_xml) == 2
    assert graphics_xml[0].get("url") == graphic_urls[0]
    assert graphics_xml[1].get("url") == graphic_urls[1]


def test_has_correct_single_graphic_url():
    graphic_urls = "Figure 1.jgp"
    charter = Charter(id_text="1", graphic_urls=graphic_urls)
    assert charter.graphic_urls[0] == graphic_urls
    graphics_xml = xps(
        charter, "/cei:text/cei:body/cei:chDesc/cei:witnessOrig/cei:figure/cei:graphic"
    )
    assert graphics_xml.get("url") == graphic_urls


def test_has_empty_graphic_urls_for_empty_text():
    graphic_urls = ""
    charter = Charter(id_text="1", graphic_urls=graphic_urls)
    assert len(charter.graphic_urls) == 0


# --------------------------------------------------------------------#
#                         Charter footnotes                          #
# --------------------------------------------------------------------#


def test_has_correct_footnotes():
    footnotes = ["Footnote a", "Footnote b"]
    charter = Charter(id_text="1", footnotes=footnotes)
    assert charter._footnotes == footnotes
    notes = xp(
        charter,
        "/cei:text/cei:back/cei:divNotes/cei:note",
    )
    assert len(notes) == 2
    assert notes[0].text == footnotes[0]
    assert notes[1].text == footnotes[1]


def test_has_no_footnotes_for_empty_text():
    footnotes = ""
    charter = Charter(id_text="1", footnotes=footnotes)
    assert len(charter._footnotes) == 0


# --------------------------------------------------------------------#
#                             Charter id                             #
# --------------------------------------------------------------------#


def test_has_correct_id():
    id_text = "~!1307 II 22|23.Ⅱ"
    id_norm = "~%211307%20II%2022%7C23.%E2%85%A1"
    charter = Charter(id_text=id_text)
    assert charter.id_text == id_text
    assert charter.id_norm == id_norm
    idno = xps(charter, "/cei:text/cei:body/cei:idno")
    assert idno.get("id") == id_norm
    assert idno.text == id_text


def test_has_correct_id_norm():
    id_text = "1307 Ⅱ 22"
    id_norm = "1307_%E2%85%A1_22"
    charter = Charter(id_text=id_text, id_norm="1307_Ⅱ_22")
    assert charter.id_text == id_text
    assert charter.id_norm == id_norm
    idno = xps(charter, "/cei:text/cei:body/cei:idno")
    assert idno.get("id") == id_norm
    assert idno.text == id_text


def test_has_correct_id_norm_for_empty_text():
    id = "1307-12-01"
    charter = Charter(id_text=id, id_norm="")
    assert charter.id_norm == id


def test_has_correct_id_old():
    id_old = "123456 α"
    idno = xps(
        Charter(id_text="1307 II 22", id_old=id_old), "/cei:text/cei:body/cei:idno"
    )
    assert idno.get("old") == id_old


def test_has_no_id_old_for_empty_text():
    charter = Charter(id_text="1307-12-01", id_old="")
    assert charter.id_old == None


def test_raises_exception_for_missing_id():
    with pytest.raises(ValueError):
        Charter(id_text="")


# --------------------------------------------------------------------#
#                       Charter index                                #
# --------------------------------------------------------------------#


def test_has_correct_index_geo_features():
    index_geo_features = [
        "Geo feature a",
        CEI.geogName("Geo feature b"),
    ]
    charter = Charter(id_text="1", index_geo_features=index_geo_features)
    assert charter.index_geo_features == index_geo_features
    geog_names_xml = xp(charter, "/cei:text/cei:back/cei:geogName")
    assert len(geog_names_xml) == 2
    assert geog_names_xml[0].text == index_geo_features[0]
    assert geog_names_xml[1].text == index_geo_features[1].text


def test_raises_exception_for_invalid_index_geo_features_xml():
    with pytest.raises(ValueError):
        Charter(
            id_text="1",
            index_geo_features=[
                CEI.persName("A person"),
                CEI.geogName("An geog name"),
            ],
        )


def test_has_correct_index_terms():
    index = [
        "Term a",
        CEI.index("Term b"),
    ]
    charter = Charter(id_text="1", index=index)
    assert charter.index == index
    index_xml = xp(charter, "/cei:text/cei:back/cei:index")
    assert len(index_xml) == 2
    assert index_xml[0].text == index[0]
    assert index_xml[1].text == index[1].text


def test_raises_exception_for_invalid_index_terms_xml():
    with pytest.raises(ValueError):
        Charter(
            id_text="1",
            index=[CEI.persName("A person"), CEI.index("An index term")],
        )


def test_has_correct_index_organizations():
    index_organizations = [
        "Organization a",
        CEI.orgName("Organization b"),
    ]
    charter = Charter(id_text="1", index_organizations=index_organizations)
    assert charter.index_organizations == index_organizations
    organization_names_xml = xp(charter, "/cei:text/cei:back/cei:orgName")
    assert len(organization_names_xml) == 2
    assert organization_names_xml[0].text == index_organizations[0]
    assert organization_names_xml[1].text == index_organizations[1].text


def test_raises_exception_for_invalid_index_organizations_xml():
    with pytest.raises(ValueError):
        Charter(
            id_text="1",
            index_organizations=[
                CEI.persName("A person"),
                CEI.orgName("An organization"),
            ],
        )


def test_has_correct_index_persons():
    index_persons = [
        "Person a",
        CEI.persName("Person b"),
        CEI.persName("Person c", {"type": "Custom Type"}),
    ]
    charter = Charter(id_text="1", index_persons=index_persons)
    assert charter.index_persons == index_persons
    pers_names_xml = xp(charter, "/cei:text/cei:back/cei:persName")
    assert len(pers_names_xml) == 3
    assert pers_names_xml[0].text == index_persons[0]
    assert pers_names_xml[1].text == index_persons[1].text
    assert pers_names_xml[2].text == index_persons[2].text
    assert pers_names_xml[2].get("type") == index_persons[2].get("type")


def test_raises_exception_for_invalid_index_persons_xml():
    with pytest.raises(ValueError):
        Charter(
            id_text="1",
            index_persons=[CEI.persName("A person"), CEI.placeName("A place")],
        )


def test_has_correct_index_places():
    index_places = [
        "Place a",
        CEI.placeName("Place b"),
    ]
    charter = Charter(id_text="1", index_places=index_places)
    assert charter.index_places == index_places
    place_names_xml = xp(charter, "/cei:text/cei:back/cei:placeName")
    assert len(place_names_xml) == 2
    assert place_names_xml[0].text == index_places[0]
    assert place_names_xml[1].text == index_places[1].text


def test_raises_exception_for_invalid_index_places_xml():
    with pytest.raises(ValueError):
        Charter(
            id_text="1",
            index_places=[CEI.persName("A person"), CEI.placeName("A place")],
        )


# --------------------------------------------------------------------#
#                        Charter issued place                        #
# --------------------------------------------------------------------#


def test_has_correct_text_issued_place():
    issued_place = "Wien"
    charter = Charter(id_text="1", issued_place=issued_place)
    assert charter.issued_place == issued_place
    issued_place_xml = xps(
        charter, "/cei:text/cei:body/cei:chDesc/cei:issued/cei:placeName"
    )
    assert issued_place_xml.text == issued_place


def test_has_no_issued_place_for_empty_text():
    issued_place = ""
    charter = Charter(id_text="1", issued_place=issued_place)
    assert charter.issued_place == None


def test_has_correct_xml_issued_place():
    issued_place = CEI.placeName("Wien")
    charter = Charter(id_text="1", issued_place=issued_place)
    assert charter.issued_place == issued_place
    issued_place_xml = xps(
        charter, "/cei:text/cei:body/cei:chDesc/cei:issued/cei:placeName"
    )
    assert issued_place_xml.text == issued_place.text


def test_raises_exception_for_incorrect_xml_issued_place():
    incorrect_element = CEI.persName("A person")
    with pytest.raises(ValueError):
        Charter(id_text="1", issued_place=incorrect_element)


# --------------------------------------------------------------------#
#                           Charter issuer                           #
# --------------------------------------------------------------------#


def test_has_correct_abstract_with_text_issuer():
    abstract = (
        "Konrad von Lintz, Caplan zu St. Pankraz, beurkundet den vorstehenden Vertrag."
    )
    issuer = "Konrad von Lintz"
    charter = Charter(id_text="1", abstract=abstract, issuer=issuer)
    assert isinstance(charter.issuer, str)
    assert charter.issuer == issuer
    issuer_xml = xps(charter, "/cei:text/cei:body/cei:chDesc/cei:abstract/cei:issuer")
    assert issuer_xml.text == issuer


def test_has_correct_abstract_with_empty_issuer():
    abstract = (
        "Konrad von Lintz, Caplan zu St. Pankraz, beurkundet den vorstehenden Vertrag."
    )
    issuer = ""
    charter = Charter(id_text="1", abstract=abstract, issuer=issuer)
    assert charter.issuer == None


def test_has_correct_abstract_with_xml_issuer():
    abstract = (
        "Konrad von Lintz, Caplan zu St. Pankraz, beurkundet den vorstehenden Vertrag."
    )
    issuer = CEI.issuer("Konrad von Lintz")
    assert isinstance(issuer, etree._Element)
    charter = Charter(id_text="1", abstract=abstract, issuer=issuer)
    assert isinstance(charter.issuer, etree._Element)
    assert charter.issuer.text == issuer.text
    issuer_xml = xps(charter, "/cei:text/cei:body/cei:chDesc/cei:abstract/cei:issuer")
    assert issuer_xml.text == issuer.text


def test_raises_exception_for_incorrect_xml_issuer():
    incorrect_element = CEI.persName("A person")
    with pytest.raises(ValueError):
        Charter(id_text="1", issuer=incorrect_element)


def test_raises_exception_for_xml_abstract_with_issuer():
    with pytest.raises(ValueError):
        Charter(id_text="1", abstract=CEI.abstract("An abstract"), issuer="An issuer")


# --------------------------------------------------------------------#
#                          Charter language                          #
# --------------------------------------------------------------------#


def test_has_correct_language():
    language = "A language"
    charter = Charter(id_text="1", language=language)
    assert charter.language == language
    language_xml = xps(
        charter,
        "/cei:text/cei:body/cei:chDesc/cei:lang_MOM",
    )
    assert language_xml.text == language


def test_has_no_language_for_empty_string():
    language = ""
    charter = Charter(id_text="1", language=language)
    assert charter.language == None


# --------------------------------------------------------------------#
#                      Charter literature lists                      #
# --------------------------------------------------------------------#


def test_has_correct_literature():
    literature = ["Entry 1", "Entry 2"]
    charter = Charter(id_text="1", literature=literature)
    assert charter.literature == literature
    literature_xml = xp(
        charter,
        "/cei:text/cei:body/cei:chDesc/cei:diplomaticAnalysis/cei:listBibl/cei:bibl",
    )
    assert len(literature_xml) == 2
    assert literature_xml[0].text == literature[0]
    assert literature_xml[1].text == literature[1]


def test_has_no_literature_for_empty_text():
    literature = ""
    charter = Charter(id_text="1", literature=literature)
    assert len(charter.literature) == 0


def test_has_correct_literature_abstracts():
    literature_abstracts = ["Entry 1", "Entry 2"]
    charter = Charter(id_text="1", literature_abstracts=literature_abstracts)
    assert charter.literature_abstracts == literature_abstracts
    literature_abstracts_xml = xp(
        charter,
        "/cei:text/cei:body/cei:chDesc/cei:diplomaticAnalysis/cei:listBiblRegest/cei:bibl",
    )
    assert len(literature_abstracts_xml) == 2
    assert literature_abstracts_xml[0].text == literature_abstracts[0]
    assert literature_abstracts_xml[1].text == literature_abstracts[1]


def test_has_no_literature_abstracts_for_empty_text():
    literature_abstracts = ""
    charter = Charter(id_text="1", literature_abstracts=literature_abstracts)
    assert len(charter.literature_abstracts) == 0


def test_has_correct_literature_depictions():
    literature_depictions = ["Entry 1", "Entry 2"]
    charter = Charter(id_text="1", literature_depictions=literature_depictions)
    assert charter.literature_depictions == literature_depictions
    literature_depictions_xml = xp(
        charter,
        "/cei:text/cei:body/cei:chDesc/cei:diplomaticAnalysis/cei:listBiblFaksimile/cei:bibl",
    )
    assert len(literature_depictions_xml) == 2
    assert literature_depictions_xml[0].text == literature_depictions[0]
    assert literature_depictions_xml[1].text == literature_depictions[1]


def test_has_no_literature_depictions_for_empty_text():
    literature_depictions = ""
    charter = Charter(id_text="1", literature_depictions=literature_depictions)
    assert len(charter.literature_depictions) == 0


def test_has_correct_literature_editions():
    literature_editions = ["Entry 1", "Entry 2"]
    charter = Charter(id_text="1", literature_editions=literature_editions)
    assert charter.literature_editions == literature_editions
    literature_editions_xml = xp(
        charter,
        "/cei:text/cei:body/cei:chDesc/cei:diplomaticAnalysis/cei:listBiblEdition/cei:bibl",
    )
    assert len(literature_editions_xml) == 2
    assert literature_editions_xml[0].text == literature_editions[0]
    assert literature_editions_xml[1].text == literature_editions[1]


def test_has_no_literature_editions_for_empty_text():
    literature_editions = ""
    charter = Charter(id_text="1", literature_editions=literature_editions)
    assert len(charter.literature_editions) == 0


def test_has_correct_literature_secondary():
    literature_secondary = ["Entry 1", "Entry 2"]
    charter = Charter(id_text="1", literature_secondary=literature_secondary)
    assert charter.literature_secondary == literature_secondary
    literature_secondary_xml = xp(
        charter,
        "/cei:text/cei:body/cei:chDesc/cei:diplomaticAnalysis/cei:listBiblErw/cei:bibl",
    )
    assert len(literature_secondary_xml) == 2
    assert literature_secondary_xml[0].text == literature_secondary[0]
    assert literature_secondary_xml[1].text == literature_secondary[1]


def test_has_no_literature_secondary_for_empty_text():
    literature_secondary = ""
    charter = Charter(id_text="1", literature_secondary=literature_secondary)
    assert len(charter.literature_secondary) == 0


# --------------------------------------------------------------------#
#                          Charter material                          #
# --------------------------------------------------------------------#


def test_has_correct_material():
    material = "Material"
    charter = Charter(id_text="1", material=material)
    assert charter.material == material
    material_xml = xps(
        charter,
        "/cei:text/cei:body/cei:chDesc/cei:witnessOrig/cei:physicalDesc/cei:material",
    )
    assert material_xml.text == material


def test_has_no_material_for_empty_text():
    material = ""
    charter = Charter(id_text="1", material=material)
    assert charter.material == None


# --------------------------------------------------------------------#
#                        Charter notarial authentication             #
# --------------------------------------------------------------------#


def test_has_correct_text_notarial_authentication():
    notarial_authentication = "A notarial authentication"
    charter = Charter(id_text="1", notarial_authentication=notarial_authentication)
    assert charter.notarial_authentication == notarial_authentication
    notarial_authentication_xml = xps(
        charter,
        "/cei:text/cei:body/cei:chDesc/cei:witnessOrig/cei:auth/cei:notariusDesc",
    )
    assert notarial_authentication_xml.text == notarial_authentication


def test_has_no_text_notarial_authentication_for_empty_text():
    notarial_authentication = ""
    charter = Charter(id_text="1", notarial_authentication=notarial_authentication)
    assert charter.notarial_authentication == None


def test_has_correct_xml_notarial_authentication():
    notarial_authentication = CEI.notariusDesc("An xml notarial authentication")
    charter = Charter(id_text="1", notarial_authentication=notarial_authentication)
    assert charter.notarial_authentication == notarial_authentication
    notarial_authentication_xml = xps(
        charter,
        "/cei:text/cei:body/cei:chDesc/cei:witnessOrig/cei:auth/cei:notariusDesc",
    )
    assert notarial_authentication_xml.text == notarial_authentication.text


def test_raises_exception_for_incorrect_xml_notarial_authentication():
    incorrect_element = CEI.persName("A person")
    with pytest.raises(ValueError):
        Charter(id_text="1", notarial_authentication=incorrect_element)


# --------------------------------------------------------------------#
#                         Charter recipient                          #
# --------------------------------------------------------------------#


def test_has_correct_abstract_with_text_recipient():
    abstract = (
        "Konrad von Lintz, Caplan zu St. Pankraz, beurkundet den vorstehenden Vertrag."
    )
    recipient = "Heinrich Müller"
    charter = Charter(id_text="1", abstract=abstract, recipient=recipient)
    assert isinstance(charter.recipient, str)
    assert charter.recipient == recipient
    recipient_xml = xps(
        charter, "/cei:text/cei:body/cei:chDesc/cei:abstract/cei:recipient"
    )
    assert recipient_xml.text == recipient


def test_has_abstract_without_text_recipient_for_empty_text():
    abstract = (
        "Konrad von Lintz, Caplan zu St. Pankraz, beurkundet den vorstehenden Vertrag."
    )
    recipient = ""
    charter = Charter(id_text="1", abstract=abstract, recipient=recipient)
    assert charter.recipient == None


def test_has_correct_abstract_with_xml_recipient():
    abstract = (
        "Konrad von Lintz, Caplan zu St. Pankraz, beurkundet den vorstehenden Vertrag."
    )
    recipient = CEI.recipient("Heinrich Müller")
    assert isinstance(recipient, etree._Element)
    charter = Charter(id_text="1", abstract=abstract, recipient=recipient)
    assert isinstance(charter.recipient, etree._Element)
    assert charter.recipient.text == recipient.text
    recipient_xml = xps(
        charter, "/cei:text/cei:body/cei:chDesc/cei:abstract/cei:recipient"
    )
    assert recipient_xml.text == recipient.text


def test_raises_exception_for_xml_abstract_with_recipient():
    with pytest.raises(ValueError):
        Charter(
            id_text="1", abstract=CEI.abstract("An abstract"), recipient="An recipient"
        )


def test_raises_exception_for_incorrect_xml_recipient():
    incorrect_element = CEI.issuer("A person")
    with pytest.raises(ValueError):
        Charter(id_text="1", recipient=incorrect_element)


# --------------------------------------------------------------------#
#                     Charter seal descriptions                      #
# --------------------------------------------------------------------#


def test_has_correct_seal_description_xml():
    seals = CEI.sealDesc(CEI.seal("Seal 1"), CEI.seal("Seal 2"))
    charter = Charter(id_text="1", seals=seals)
    assert charter.seals == seals
    seals_xml = xp(
        charter,
        "/cei:text/cei:body/cei:chDesc/cei:witnessOrig/cei:auth/cei:sealDesc/cei:seal",
    )
    assert len(seals_xml) == 2
    assert seals_xml[0].text == "Seal 1"
    assert seals_xml[1].text == "Seal 2"


def test_has_no_seal_description_for_empty_text():
    charter = Charter(id_text="1", seals="")
    assert charter.seals == None


def test_has_correct_seal_text_description():
    seals = "2 Siegel"
    charter = Charter(id_text="1", seals=seals)
    assert charter.seals == seals
    seals_xml = xps(
        charter, "/cei:text/cei:body/cei:chDesc/cei:witnessOrig/cei:auth/cei:sealDesc"
    )
    assert seals_xml.text == seals


def test_has_correct_multiple_seal_text_descriptions():
    seals = ["Seal 1", "Seal 2"]
    charter = Charter(id_text="1", seals=seals)
    assert charter.seals == seals
    seals_xml = xp(
        charter,
        "/cei:text/cei:body/cei:chDesc/cei:witnessOrig/cei:auth/cei:sealDesc/cei:seal",
    )
    assert len(seals_xml) == 2
    assert seals_xml[0].text == "Seal 1"
    assert seals_xml[1].text == "Seal 2"


def test_has_correct_single_seal_object():
    seals = Seal(material="A material", sigillant="A sigillant")
    charter = Charter(id_text="1", seals=seals)
    assert charter.seals == seals
    seals_xml = xps(
        charter,
        "/cei:text/cei:body/cei:chDesc/cei:witnessOrig/cei:auth/cei:sealDesc/cei:seal",
    )
    assert seals_xml.xpath("cei:sealMaterial/text()", namespaces=CHARTER_NSS) == [
        "A material"
    ]
    assert seals_xml.xpath("cei:sigillant/text()", namespaces=CHARTER_NSS) == [
        "A sigillant"
    ]


def test_has_correct_multiple_seal_objects():
    seals = [
        Seal(material="Material a", sigillant="Sigillant a"),
        Seal(material="Material b", sigillant="Sigillant b"),
    ]
    charter = Charter(id_text="1", seals=seals)
    assert charter.seals == seals
    seals_xml = xp(
        charter,
        "/cei:text/cei:body/cei:chDesc/cei:witnessOrig/cei:auth/cei:sealDesc/cei:seal",
    )
    assert seals_xml[0].xpath("cei:sealMaterial/text()", namespaces=CHARTER_NSS) == [
        "Material a"
    ]
    assert seals_xml[0].xpath("cei:sigillant/text()", namespaces=CHARTER_NSS) == [
        "Sigillant a"
    ]
    assert seals_xml[1].xpath("cei:sealMaterial/text()", namespaces=CHARTER_NSS) == [
        "Material b"
    ]
    assert seals_xml[1].xpath("cei:sigillant/text()", namespaces=CHARTER_NSS) == [
        "Sigillant b"
    ]


# --------------------------------------------------------------------#
#                       Charter transcription                        #
# --------------------------------------------------------------------#


def test_has_correct_text_transcription():
    transcription = "A transcription"
    charter = Charter(id_text="1", transcription=transcription)
    assert charter.transcription == transcription
    transcription_xml = xps(charter, "/cei:text/cei:body/cei:tenor")
    assert transcription_xml.text == transcription


def test_has_no_text_transcription_for_empty_text():
    transcription = ""
    charter = Charter(id_text="1", transcription=transcription)
    assert charter.transcription == None


def test_has_correct_xml_transcription():
    transcription = CEI.tenor("Tenor with ", CEI.sup("a"), " superscript")
    charter = Charter(id_text="1", transcription=transcription)
    assert charter.transcription == transcription
    superscript_xml = xps(
        charter,
        "/cei:text/cei:body/cei:tenor/cei:sup",
    )
    assert superscript_xml.text == "a"


def test_raises_exception_for_incorrect_xml_transcription():
    incorrect_element = CEI.persName("A person")
    with pytest.raises(ValueError):
        Charter(id_text="1", transcription=incorrect_element)


# --------------------------------------------------------------------#
#                       Charter tradition form                       #
# --------------------------------------------------------------------#


def test_has_correct_tradition():
    tradition = "orig."
    charter = Charter(id_text="1", tradition=tradition)
    assert charter.tradition == tradition
    tradition_xml = xps(
        charter, "/cei:text/cei:body/cei:chDesc/cei:witnessOrig/cei:traditioForm"
    )
    assert tradition_xml.text == tradition


def test_has_no_tradition_for_empty_text():
    tradition = ""
    charter = Charter(id_text="1", tradition=tradition)
    assert charter.tradition == None


# --------------------------------------------------------------------#
#                            Charter type                            #
# --------------------------------------------------------------------#


def test_has_correct_type():
    charter = Charter(id_text="1").to_xml()
    assert charter.get("type") == "charter"


# --------------------------------------------------------------------#
#                         Charter witnesses                          #
# --------------------------------------------------------------------#


def test_has_correct_witnesses():
    witnesses = [
        "Witness a",
        CEI.persName("Witness b"),
        CEI.persName("Witness c", {"type": "Zeuge"}),
    ]
    charter = Charter(id_text="1", witnesses=witnesses)
    assert charter.witnesses == witnesses
    pers_names_xml = xp(charter, '/cei:text/cei:back/cei:persName[@type="Zeuge"]')
    assert len(pers_names_xml) == 3
    assert pers_names_xml[0].text == witnesses[0]
    assert pers_names_xml[1].text == witnesses[1].text
    assert pers_names_xml[2].text == witnesses[2].text


def test_raises_exception_for_invalid_witnesses_xml():
    with pytest.raises(ValueError):
        Charter(
            id_text="1", witnesses=[CEI.persName("A Person"), CEI.placeName("A place")]
        )



from typing import List, Optional

from lxml import etree

from to_cei.charter import Charter
from to_cei.config import CEI, CEI_SCHEMA_LOCATION_ATTRIBUTE
from to_cei.xml_assembler import XmlAssembler


class CharterGroup(XmlAssembler):
    _charters: List[Charter] = []
    _name: str = ""

    def __init__(self, name: str, charters: List[Charter] = []):
        """Creates a new charter group object.

        Args:
            name (str): The name of the charter group. Is not allowed to be empty
            charters(List[Charter] = []): An optional list of Charter objects
        """
        self.name = name
        self.charters = charters

    @property
    def charters(self):
        return self._charters

    @charters.setter
    def charters(self, value: List[Charter]):
        self._charters = value

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value: str):
        if len(value) == 0:
            raise ValueError("Group name cannot be empty")
        self._name = value

    def to_xml(self, add_schema_location: bool = False) -> etree._Element:
        """Creates an xml representation of the charter group.

        Args:
            add_schema_location: If True, the CEI schema location is added to the root element.

        Returns:
            An etree Element object representing the charter group.
        """
        cei = CEI.cei(
            CEI.teiHeader(CEI.fileDesc(CEI.titleStmt(CEI.title(self.name)))),
            CEI.text(CEI.group(*[charter.to_xml() for charter in self.charters])),
        )
        if add_schema_location:
            cei.attrib.update(CEI_SCHEMA_LOCATION_ATTRIBUTE)
        return cei

    def to_file(self, folder: Optional[str] = None, add_schema_location: bool = False):
        """Writes the xml representation of the charter group to a file. The filename is generated from a normalization of the group name.

        Args:
            folder (str): The folder to write the file to. If this is ommitted, the file is written to the place where the script is executed from.
            add_schema_location (bool): If True, the CEI schema location is added to the root element. Defaults to False.
        """
        return super(CharterGroup, self).to_file(
            self.name.lower().replace(" ", "_") + ".cei.group",
            folder=folder,
            add_schema_location=add_schema_location,
        )



from typing import List, Optional, Tuple

from lxml import etree

from to_cei.config import CEI
from to_cei.helpers import get_str, get_str_or_element
from to_cei.xml_assembler import XmlAssembler


class Seal(XmlAssembler):
    _condition: Optional[str] = None
    _dimensions: Optional[str] = None
    _legend: Optional[str] | List[Tuple[str, str]] = []
    _material: Optional[str] = None
    _sigillant: Optional[str | etree._Element] = None

    def __init__(
        self,
        condition: Optional[str] = None,
        dimensions: Optional[str] = None,
        legend: Optional[str | List[Tuple[str, str]]] = [],
        material: Optional[str] = None,
        sigillant: Optional[str | etree._Element] = None,
    ) -> None:
        """
        Creates a seal instance.

        Parameters:
        -----------
        condition: A description of the seal's condition.

        dimensions: A description of the seal's dimensions.

        legend: The seals legend or legends. Can bei either a text or a list of tuples of two strings where the first one is the place of the legend and the second the legend itselfself.

        material: A description of the seal's material.

        sigillant: The sigillant of the seal, either as a text or an etree._Element that contains either a cei:persName or a cei:orgName.
        """
        self.condition = condition
        self.dimensions = dimensions
        self.legend = legend
        self.material = material
        self.sigillant = sigillant

    # --------------------------------------------------------------------#
    #                             Properties                             #
    # --------------------------------------------------------------------#

    @property
    def condition(self):
        return self._condition

    @condition.setter
    def condition(self, value: Optional[str] = None):
        self._condition = get_str(value)

    @property
    def dimensions(self):
        return self._dimensions

    @dimensions.setter
    def dimensions(self, value: Optional[str] = None):
        self._dimensions = get_str(value)

    @property
    def legend(self):
        return self._legend

    @legend.setter
    def legend(self, value: Optional[str | List[Tuple[str, str]]] = []):
        self._legend = (
            []
            if value is None or (isinstance(value, str) and not len(value))
            else value
        )

    @property
    def material(self):
        return self._material

    @material.setter
    def material(self, value: Optional[str] = None):
        self._material = get_str(value)

    @property
    def sigillant(self):
        return self._sigillant

    @sigillant.setter
    def sigillant(self, value: Optional[str | etree._Element] = None):
        self._sigillant = get_str_or_element(value, "persName", "orgName")

    # --------------------------------------------------------------------#
    #                           Public methods                           #
    # --------------------------------------------------------------------#

    def to_xml(self) -> Optional[etree._Element]:
        children = []
        if self.condition is not None:
            children.append(CEI.sealCondition(self.condition))
        if self.dimensions is not None:
            children.append(CEI.sealDimensions(self.dimensions))
        if isinstance(self.legend, str):
            children.append(CEI.legend(self.legend))
        if isinstance(self.legend, List):
            children = children + [
                CEI.legend(legend[1], {"place": legend[0]}) for legend in self.legend
            ]
        if self.material is not None:
            children.append(CEI.sealMaterial(self.material))
        if self.sigillant is not None:
            children.append(CEI.sigillant(self.sigillant))
        return CEI.seal(*children) if len(children) else None



import re
from datetime import datetime
from typing import List, Optional, Tuple
from urllib.parse import quote

from astropy.time import Time
from lxml import etree

from to_cei.config import CEI, CEI_SCHEMA_LOCATION_ATTRIBUTE
from to_cei.helpers import (get_str, get_str_list, get_str_or_element,
                            get_str_or_element_list, join)
from to_cei.seal import Seal
from to_cei.xml_assembler import XmlAssembler

MOM_DATE_REGEX = re.compile(
    r"^(?P<year>-?[129]?[0-9][0-9][0-9])(?P<month>[019][0-9])(?P<day>[01239][0-9])$"
)
NO_DATE_TEXT = "No date"
NO_DATE_VALUE = "99999999"

SIMPLE_URL_REGEX = re.compile(r"^https?://.{1,}\..{1,}$")

Date = str | datetime | Time

DateValue = Optional[Date | Tuple[Date, Date]]


def to_mom_date_value(time: Time) -> str:
    """Converts an astropy.Time object to a mom-compatible date string.

    Args:
        time (Time): An astropy.Time object

    Returns:
        A date string compatible with mom-ca.
    """
    year = time.ymdhms[0]
    month = time.ymdhms[1]
    day = time.ymdhms[2]
    return "{year}{month}{day}".format(
        year=str(year).zfill(3) if year >= 0 else "-" + str(year * -1).zfill(3),
        month=str(month).zfill(2),
        day=str(day).zfill(2),
    )


def mom_date_to_time(value: str) -> Time:
    """Converts a mom-compatible date string into an astropy.Time object if possible.

    Args:
        value (str): A mom-compatible date string in the form of -?[129]?[0-9][0-9][0-9][019][0-9][01239][0-9]

    Returns:
        A astropy.Time object

    Raises:
        ValueError: If the provided value cannot be converted to a valid astropy.Time object

    """
    match = re.search(MOM_DATE_REGEX, value)
    if match is None:
        raise ValueError("Invalid mom date value provided: '{}'".format(value))
    year = match.group("year")
    if not isinstance(year, str):
        raise ValueError("Invalid year in mom date value: {}".format(year))
    month = match.group("month")
    if not isinstance(month, str):
        raise ValueError("Invalid month in mom date value: {}".format(month))
    day = match.group("day")
    if not isinstance(day, str):
        raise ValueError("Invalid day in mom date value: {}".format(day))
    return Time(
        {"year": int(year), "month": int(month), "day": int(day)},
        format="ymdhms",
        scale="ut1",
    )


def string_to_time(value: str | Tuple[str, str]) -> Time | Tuple[Time, Time]:
    """Converts a single date string or a tuple of date strings to a matching single or tuple astropy.Time object.

    Args:
        value (str | Tuple[str, str]): A single or tuple of date strings. Can be either an iso-compatible or a mom-ca compatible date string.

    Returns:
        A single or tuple astropy.Time objectself.

    Raises:
        ValueError: If the date/s cannot be converted to astropy.Time objects.
    """
    if isinstance(value, Tuple) and len(value) != 2:
        raise ValueError("Invalid date tuple provided: '{}'".format(value))
    try:
        # Try to directly convert from an iso date string
        return (
            Time(value, format="isot", scale="ut1")
            if isinstance(value, str)
            else (
                Time(value[0], format="isot", scale="ut1"),
                Time(value[1], format="isot", scale="ut1"),
            )
        )
    # Direct conversion not possible, try to convert mom date strings
    except Exception:
        if isinstance(value, Tuple):
            return (mom_date_to_time(value[0]), mom_date_to_time(value[1]))
        else:
            return mom_date_to_time(value)


class Charter(XmlAssembler):
    _abstract: Optional[str | etree._Element] = None
    _abstract_sources: List[str] = []
    _archive: Optional[str] = None
    _chancellary_remarks: List[str] = []
    _comments: List[str] = []
    _condition: Optional[str] = None
    _date: Optional[str | etree._Element] = None
    _date_quote: Optional[str | etree._Element] = None
    _date_value: Optional[Time | Tuple[Time, Time]] = None
    _dimensions: Optional[str] = None
    _external_link: Optional[str] = None
    _footnotes: List[str] = []
    _graphic_urls: List[str] = []
    _id_norm: Optional[str] = None
    _id_old: Optional[str] = None
    _id_text: str = ""
    _index: List[str | etree._Element] = []
    _index_geo_features: List[str | etree._Element] = []
    _index_organizations: List[str | etree._Element] = []
    _index_persons: List[str | etree._Element] = []
    _index_places: List[str | etree._Element] = []
    _issued_place: Optional[str | etree._Element] = None
    _issuer: Optional[str | etree._Element] = None
    _language: Optional[str] = None
    _literature: List[str] = []
    _literature_abstracts: List[str] = []
    _literature_depictions: List[str] = []
    _literature_editions: List[str] = []
    _literature_secondary: List[str] = []
    _material: Optional[str] = None
    _notarial_authentication: Optional[str | etree._Element] = None
    _recipient: Optional[str | etree._Element] = None
    _seals: Optional[etree._Element | str | Seal | List[str] | List[Seal]] = None
    _tradition: Optional[str] = None
    _transcription: Optional[str | etree._Element] = None
    _transcription_sources: List[str] = []
    _witnesses: List[str | etree._Element] = []

    def __init__(
        self,
        id_text: str,
        abstract: Optional[str | etree._Element] = None,
        abstract_sources: Optional[str | List[str]] = [],
        archive: Optional[str] = None,
        chancellary_remarks: Optional[str | List[str]] = [],
        comments: Optional[str | List[str]] = [],
        condition: Optional[str] = None,
        date: Optional[str | etree._Element] = None,
        date_quote: Optional[str | etree._Element] = None,
        date_value: Optional[DateValue] = None,
        dimensions: Optional[str] = None,
        external_link: Optional[str] = None,
        footnotes: Optional[str | List[str]] = [],
        graphic_urls: Optional[str | List[str]] = [],
        id_norm: Optional[str] = None,
        id_old: Optional[str] = None,
        index: Optional[List[str | etree._Element]] = [],
        index_geo_features: Optional[List[str | etree._Element]] = [],
        index_organizations: Optional[List[str | etree._Element]] = [],
        index_persons: Optional[List[str | etree._Element]] = [],
        index_places: Optional[List[str | etree._Element]] = [],
        issued_place: Optional[str | etree._Element] = None,
        issuer: Optional[str | etree._Element] = None,
        language: Optional[str] = None,
        literature: Optional[str | List[str]] = [],
        literature_abstracts: Optional[str | List[str]] = [],
        literature_depictions: Optional[str | List[str]] = [],
        literature_editions: Optional[str | List[str]] = [],
        literature_secondary: Optional[str | List[str]] = [],
        material: Optional[str] = None,
        notarial_authentication: Optional[str | etree._Element] = None,
        recipient: Optional[str | etree._Element] = None,
        seals: Optional[etree._Element | str | Seal | List[str] | List[Seal]] = None,
        tradition: Optional[str] = None,
        transcription: Optional[str | etree._Element] = None,
        transcription_sources: Optional[str] | List[str] = [],
        witnesses: Optional[List[str | etree._Element]] = [],
    ) -> None:
        """
        Creates a new charter object. Empty strings in the parameters are treated similar to None values.

        Args:
            id_text: The human readable id of the charter. If id_norm is not present, id_text will be used in a normalized form. If it is missing or empty, an exception will be raised.
            abstract: The abstract either as a simple text or a complete cei:abstract etree._Element.
            abstract_sources: The bibliography source or sources for the abstract.
            archive: The name of the archive that owns the original charter.
            chancellary_remarks: Chancellary remarks as a single text or list of texts.
            comments: Diplomatic commentary as text or list of texts.
            condition: A description of the charter's condition in text form.
            date: The date the charter was issued at either as text to use when converting to CEI or a complete cei:date or cei:dateRange etree._Element. If the date is given as an XML element, date_value needs to remain emptyself. Missing values will be constructed as having a date of "No date" in the XML.
            date_quote: The charter's date in the original text either as text or a complete cel:quoteOriginaldatierung etree._Element object.
            date_value: The actual date value in case the value in date is just a text and not an XML element. Can bei either an ISO date string, a MOM-compatible string, a python datetime object (can only be between years 1 and 9999) or an astropy Time object - or a tuple with two such values. If a single value is given, it is interpreted as an exact value, otherwise the two values will be used as from/to attributes of a cei:dateRange object. Missing values will be added to the xml as @value="99999999" to conform with the MOM data practices. When a date_value is added and date is an XML element, an exception is raised.
            dimensions: The description of the physical dimensions of the charter as text.
            external_link: A link to an external representation of the charter as text.
            footnotes: Footnotes as text or list of texts.
            graphic_urls: A list of strings that represents the urls of various images representing the charter. Can bei either full urls or just the filenames of the image files, depending on the charter fond / collection settings.
            id_norm: A normalized id for the charter. It will be percent-encoded to ensure only valid characters are used. If it is ommitted, the normalized version of id_text will be used.
            id_old: An old, now obsolete identifier of the charter.
            index: A list of terms as texts or cei:index etree._Element objects to be included in the index.
            index_geo_features: A list of geographical features as texts or cei:geogName etree._Element objects to be included in the index.
            index_organizations: A list of organizations as texts or cei:orgName etree._Element objects to be included in the index.
            index_persons: A list of persons as texts or cei:persName etree._Element objects to be included in the index.
            index_places: A list of places as texts or cei:placeName etree._Element objects to be included in the index.
            issued_place: The place the charter has been issued at either as text or a complete cei:placeName etree._Element.
            issuer: The issuer of the charter either as text or a complete cei:issuer etree._Element.
            language: The language of the charter as text.
            literature: A single text or list of texts descibing unspecified literature for the charter.
            literature_abstracts: A single text or list of texts descibing abstracts of the charter.
            literature_depictions: A single text or list of texts descibing depictions of the charter.
            literature_editions: A single text or list of texts descibing editions of the charter.
            literature_secondary: A single text or list of texts descibing secondary literature about the charter.
            material: A string description of the material the charter is made of.
            notarial_authentication: A string or complete cei:notariusDesc etree._Element that describes the notarial_authentication of the charter.
            recipient: The recipient of the charter either as text or a complete cei:issuer etree._Element.
            seals: The description of the seals of a charter, either as a single/list of simple text descriptions or Seal objects, or a complete cei:sealDesc etree._Element object.
            tradition: The status of tradition of the charter, as an original, copy or something else. Can be any free text.
            transcription: The full text transcription of the charter either as text or a complete cei:tenor etree._Element object.
            transcription_sources: The source or sources for the transcription.
            witnesses: The list of witnesses either as text or complete cei:persName etree._Element objects. A '@type="Zeuge"' attribute will be added for all etree._Element list items.

        Raises:
            ValueError: Raised when various values don't make sense in the context of the charter creation.
        """
        if not id_text:
            raise ValueError("id_text is not allowed to be empty")
        self.abstract = abstract
        self.abstract_sources = abstract_sources
        self.archive = archive
        self.chancellary_remarks = chancellary_remarks
        self.comments = comments
        self.condition = condition
        self.date = date
        self.date_quote = date_quote
        if date_value is not None:
            self.date_value = date_value
        self.dimensions = dimensions
        self.external_link = external_link
        self.footnotes = footnotes
        self.graphic_urls = graphic_urls
        self.id_norm = id_norm
        self.id_old = id_old
        self.id_text = id_text
        self.index = index
        self.index_geo_features = index_geo_features
        self.index_organizations = index_organizations
        self.index_persons = index_persons
        self.index_places = index_places
        self.issued_place = issued_place
        self.issuer = issuer
        self.language = language
        self.literature = literature
        self.literature_abstracts = literature_abstracts
        self.literature_depictions = literature_depictions
        self.literature_editions = literature_editions
        self.literature_secondary = literature_secondary
        self.material = material
        self.notarial_authentication = notarial_authentication
        self.recipient = recipient
        self.seals = seals
        self.tradition = tradition
        self.transcription = transcription
        self.transcription_sources = transcription_sources
        self.witnesses = witnesses

    # --------------------------------------------------------------------#
    #                             Properties                             #
    # --------------------------------------------------------------------#

    @property
    def abstract(self):
        return self._abstract

    @abstract.setter
    def abstract(self, value: Optional[str | etree._Element] = None):
        if self.issuer is not None and isinstance(self.issuer, etree._Element):
            raise ValueError(
                "XML element content for both issuer and abstract is not allowed, please join the issuer in the XML abstract yourself"
            )
        self._abstract = get_str_or_element(value, "abstract")

    @property
    def abstract_sources(self):
        return self._abstract_sources

    @abstract_sources.setter
    def abstract_sources(self, value: Optional[str | List[str]] = []):
        self._abstract_sources = get_str_list(value)

    @property
    def archive(self):
        return self._archive

    @archive.setter
    def archive(self, value: Optional[str] = None):
        self._archive = get_str(value)

    @property
    def chancellary_remarks(self):
        return self._chancellary_remarks

    @chancellary_remarks.setter
    def chancellary_remarks(self, value: Optional[str | List[str]] = []):
        self._chancellary_remarks = get_str_list(value)

    @property
    def comments(self):
        return self._comments

    @comments.setter
    def comments(self, value: Optional[str | List[str]] = []):
        self._comments = get_str_list(value)

    @property
    def condition(self):
        return self._condition

    @condition.setter
    def condition(self, value: Optional[str] = None):
        self._condition = get_str(value)

    @property
    def date(self):
        return self._date

    @date.setter
    def date(self, value: Optional[str | etree._Element] = None):
        self._date = get_str_or_element(value, "date", "dateRange")

    @property
    def date_quote(self):
        return self._date_quote

    @date_quote.setter
    def date_quote(self, value: Optional[str | etree._Element] = None):
        self._date_quote = get_str_or_element(value, "quoteOriginaldatierung")

    @property
    def date_value(self):
        return self._date_value

    @date_value.setter
    def date_value(self, value: Optional[DateValue] = None):
        # Don't allow to directly set date values if an XML date element is present
        if isinstance(self.date, etree._Element):
            raise ValueError(
                "Not allowed to set date value directly if the date is already an XML element."
            )
        # Unknown MOM date (99999999)
        elif (
            isinstance(value, str) and (value == NO_DATE_VALUE or not len(value))
        ) or (
            isinstance(value, Tuple)
            and len(value) == 2
            and value[0] == NO_DATE_VALUE
            and value[1] == NO_DATE_VALUE
        ):
            self._date_value = None
        # Directly set None, Time and [Time, Time] values
        elif (
            value is None
            or isinstance(value, Time)
            or (
                isinstance(value, Tuple)
                and len(value) == 2
                and isinstance(value[0], Time)
                and isinstance(value[1], Time)
            )
        ):
            self._date_value = value  # type: ignore
        # Convert python date objects
        elif isinstance(value, datetime):
            self._date_value = Time(value, scale="ut1")
        # Convert python date tuples
        elif (
            isinstance(value, Tuple)
            and len(value) == 2
            and isinstance(value[0], datetime)
            and isinstance(value[1], datetime)
        ):
            self._date_value = (
                Time(value[0], scale="ut1"),
                Time(value[1], scale="ut1"),
            )
        # Convert strings
        elif isinstance(value, str):
            self._date_value = string_to_time(value)
        # Convert string tuples
        elif (
            isinstance(value, Tuple)
            and len(value) == 2
            and isinstance(value[0], str)
            and isinstance(value[1], str)
        ):
            self._date_value = string_to_time(value)  # type: ignore
        else:
            raise ValueError("Invalid date value: '{}'".format(value))

    @property
    def dimensions(self):
        return self._dimensions

    @dimensions.setter
    def dimensions(self, value: Optional[str] = None):
        self._dimensions = get_str(value)

    @property
    def external_link(self):
        return self._external_link

    @external_link.setter
    def external_link(self, value: Optional[str] = None):
        if not isinstance(value, str) or len(value) == 0:
            return None
        if not re.match(SIMPLE_URL_REGEX, value):
            raise ValueError(
                "'{}' does not look like a valid external URL. If you think it is valid, please contact the to-CEI library maintainers and tell them.".format(
                    value
                )
            )
        self._external_link = value

    @property
    def footnotes(self):
        return self._footnotes

    @footnotes.setter
    def footnotes(self, value: Optional[str | List[str]] = []):
        self._footnotes = get_str_list(value)

    @property
    def graphic_urls(self):
        return self._graphic_urls

    @graphic_urls.setter
    def graphic_urls(self, value: Optional[str | List[str]] = []):
        self._graphic_urls = get_str_list(value)

    @property
    def id_norm(self):
        return quote(self._id_norm if self._id_norm else self.id_text)

    @id_norm.setter
    def id_norm(self, value: Optional[str] = None):
        self._id_norm = get_str(value)

    @property
    def id_old(self):
        return self._id_old

    @id_old.setter
    def id_old(self, value: Optional[str] = None):
        self._id_old = get_str(value)

    @property
    def id_text(self):
        return self._id_text

    @id_text.setter
    def id_text(self, value: str):
        self._id_text = value

    @property
    def index(self):
        return self._index

    @index.setter
    def index(self, value: Optional[List[str | etree._Element]] = []):
        self._index = get_str_or_element_list(value, "index")

    @property
    def index_geo_features(self):
        return self._index_geo_features

    @index_geo_features.setter
    def index_geo_features(self, value: Optional[List[str | etree._Element]] = []):
        self._index_geo_features = get_str_or_element_list(value, "geogName")

    @property
    def index_organizations(self):
        return self._index_organizations

    @index_organizations.setter
    def index_organizations(self, value: Optional[List[str | etree._Element]] = []):
        self._index_organizations = get_str_or_element_list(value, "orgName")

    @property
    def index_persons(self):
        return self._index_persons

    @index_persons.setter
    def index_persons(self, value: Optional[List[str | etree._Element]] = []):
        self._index_persons = get_str_or_element_list(value, "persName")

    @property
    def index_places(self):
        return self._index_places

    @index_places.setter
    def index_places(self, value: Optional[List[str | etree._Element]] = []):
        self._index_places = get_str_or_element_list(value, "placeName")

    @property
    def issued_place(self):
        return self._issued_place

    @issued_place.setter
    def issued_place(self, value: Optional[str | etree._Element] = None):
        self._issued_place = get_str_or_element(value, "placeName")

    @property
    def issuer(self):
        return self._issuer

    @issuer.setter
    def issuer(self, value: Optional[str | etree._Element] = None):
        if value is not None and isinstance(self.abstract, etree._Element):
            raise ValueError(
                "XML element content for both issuer and abstract is not allowed, please join the issuer in the XML abstract yourself"
            )
        self._issuer = get_str_or_element(value, "issuer")

    @property
    def language(self):
        return self._language

    @language.setter
    def language(self, value: Optional[str] = None):
        self._language = get_str(value)

    @property
    def literature(self):
        return self._literature

    @literature.setter
    def literature(self, value: Optional[str | List[str]] = []):
        self._literature = get_str_list(value)

    @property
    def literature_abstracts(self):
        return self._literature_abstracts

    @literature_abstracts.setter
    def literature_abstracts(self, value: Optional[str | List[str]] = []):
        self._literature_abstracts = get_str_list(value)

    @property
    def literature_depictions(self):
        return self._literature_depictions

    @literature_depictions.setter
    def literature_depictions(self, value: Optional[str | List[str]] = []):
        self._literature_depictions = get_str_list(value)

    @property
    def literature_editions(self):
        return self._literature_editions

    @literature_editions.setter
    def literature_editions(self, value: Optional[str | List[str]] = []):
        self._literature_editions = get_str_list(value)

    @property
    def literature_secondary(self):
        return self._literature_secondary

    @literature_secondary.setter
    def literature_secondary(self, value: Optional[str | List[str]] = []):
        self._literature_secondary = get_str_list(value)

    @property
    def material(self):
        return self._material

    @material.setter
    def material(self, value: Optional[str] = None):
        self._material = get_str(value)

    @property
    def notarial_authentication(self):
        return self._notarial_authentication

    @notarial_authentication.setter
    def notarial_authentication(self, value: Optional[str | etree._Element] = None):
        self._notarial_authentication = get_str_or_element(value, "notariusDesc")

    @property
    def recipient(self):
        return self._recipient

    @recipient.setter
    def recipient(self, value: Optional[str | etree._Element] = None):
        if value is not None and isinstance(self.abstract, etree._Element):
            raise ValueError(
                "XML element content for both recipient and abstract is not allowed, please join the recipient in the XML abstract yourself"
            )
        self._recipient = get_str_or_element(value, "recipient")

    @property
    def seals(self):
        return self._seals

    @seals.setter
    def seals(
        self,
        value: Optional[etree._Element | str | Seal | List[str] | List[Seal]] = None,
    ):
        if value is None or isinstance(value, str) and len(value) == 0:
            return None
        validated = (
            get_str_or_element(value, "sealDesc")
            if isinstance(value, etree._Element)
            else value
        )
        if validated is None:
            self._seals = None
        else:
            self._seals = validated

    @property
    def tradition(self):
        return self._tradition

    @tradition.setter
    def tradition(self, value: Optional[str] = None):
        self._tradition = get_str(value)

    @property
    def transcription(self):
        return self._transcription

    @transcription.setter
    def transcription(self, value: Optional[str | etree._Element] = None):
        self._transcription = get_str_or_element(value, "tenor")

    @property
    def transcription_sources(self):
        return self._transcription_sources

    @transcription_sources.setter
    def transcription_sources(self, value: Optional[str | List[str]] = []):
        self._transcription_sources = get_str_list(value)

    @property
    def witnesses(self):
        return self._witnesses

    @witnesses.setter
    def witnesses(self, value: Optional[List[str | etree._Element]] = []):
        self._witnesses = get_str_or_element_list(value, "persName")

    # --------------------------------------------------------------------#
    #                        Private CEI creators                        #
    # --------------------------------------------------------------------#

    def _create_cei_abstract(self) -> Optional[etree._Element]:
        children = join(self._create_cei_recipient(), self._create_cei_issuer())
        return (
            CEI.abstract(self.abstract, *children)
            if isinstance(self.abstract, str)
            else self.abstract
        )

    def _create_cei_arch(self) -> Optional[etree._Element]:
        return None if not self.archive else CEI.arch(self.archive)

    def _create_cei_arch_identifier(self) -> Optional[etree._Element]:
        children = join(self._create_cei_arch(), self._create_cei_ref())
        return CEI.archIdentifier(*children) if len(children) else None

    def _create_cei_auth(self) -> Optional[etree._Element]:
        children = join(self._create_cei_notarius_desc(), self._create_cei_seal_desc())
        return CEI.auth(*children) if len(children) else None

    def _create_cei_back(self) -> etree._Element:
        children = join(
            [self._create_cei_pers_name(person, type="Zeuge") for person in self.witnesses],  # type: ignore
            [self._create_cei_pers_name(person) for person in self.index_persons],  # type: ignore
            [self._create_cei_org_name(organization) for organization in self.index_organizations],  # type: ignore
            [self._create_cei_place_name(place) for place in self.index_places],  # type: ignore
            [self._create_cei_geog_name(geo_feature) for geo_feature in self.index_geo_features],  # type: ignore
            [self._create_cei_index(term) for term in self.index],  # type: ignore
            self._create_cei_div_notes(),
        )
        return CEI.back(*children)

    def _create_cei_bibls(self, bibls: List[str]) -> List[etree._Element]:
        return [CEI.bibl(bibl) for bibl in bibls]

    def _create_cei_body(self) -> etree._Element:
        children = join(
            self._create_cei_idno(), self._create_cei_chdesc(), self._create_cei_tenor()
        )
        return CEI.body(*children)

    def _create_cei_chdesc(self) -> Optional[etree._Element]:
        children = join(
            self._create_cei_abstract(),
            self._create_cei_issued(),
            self._create_cei_witness_orig(),
            self._create_cei_diplomatic_analysis(),
            self._create_cei_lang_mom(),
        )
        return CEI.chDesc(*children) if len(children) else None

    def _create_cei_condition(self) -> Optional[etree._Element]:
        return None if self.condition is None else CEI.condition(self.condition)

    def _create_cei_date(self) -> etree._Element:
        # An xml date
        if isinstance(self.date, etree._Element):
            return self.date
        # A date range tuple
        if isinstance(self.date_value, Tuple):
            return CEI.dateRange(
                "{} - {}".format(
                    self.date_value[0].to_value("fits", subfmt="longdate"),
                    self.date_value[1].to_value("fits", subfmt="longdate"),
                )
                if self.date is None
                else self.date,
                {
                    "from": to_mom_date_value(self.date_value[0]),
                    "to": to_mom_date_value(self.date_value[1]),
                },
            )
        # A single date value
        if isinstance(self.date_value, Time):
            return CEI.date(
                self.date_value.to_value("fits", subfmt="longdate")
                if self.date is None
                else self.date,
                {"value": to_mom_date_value(self.date_value)},
            )
        # Only a date text value
        if isinstance(self.date, str):
            return CEI.date(self.date, {"value": NO_DATE_VALUE})
        # Nothing
        return CEI.date(NO_DATE_TEXT, {"value": NO_DATE_VALUE})

    def _create_cei_dimensions(self) -> Optional[etree._Element]:
        return None if self.dimensions is None else CEI.dimensions(self.dimensions)

    def _create_cei_diplomatic_analysis(self) -> Optional[etree._Element]:
        children = join(
            self._create_cei_list_bibl(),
            self._create_cei_list_bibl_edition(),
            self._create_cei_list_bibl_regest(),
            self._create_cei_list_bibl_faksimile(),
            self._create_cei_list_bibl_erw(),
            self._create_cei_quote_originaldatierung(),
            self._create_cei_p(),
        )
        return CEI.diplomaticAnalysis(*children) if len(children) else None

    def _create_cei_div_notes(self) -> List[etree._Element]:
        return (
            CEI.divNotes(*[CEI.note(note) for note in self.footnotes])
            if len(self.footnotes)
            else []
        )

    def _create_cei_figures(self) -> List[etree._Element]:
        return (
            [CEI.figure(CEI.graphic({"url": url})) for url in self.graphic_urls]
            if len(self.graphic_urls)
            else []
        )

    def _create_cei_front(self) -> etree._Element:
        children = join(self._create_cei_source_desc())
        return CEI.front(*children)

    def _create_cei_idno(self) -> etree._Element:
        attributes = {"id": self.id_norm}
        if self.id_old:
            attributes["old"] = self.id_old
        return CEI.idno(self.id_text, **attributes)

    def _create_cei_issued(self) -> Optional[etree._Element]:
        children = join(
            self._create_cei_place_name(self.issued_place), self._create_cei_date()
        )
        return CEI.issued(*children) if len(children) else None

    def _create_cei_issuer(self) -> Optional[etree._Element]:
        return (
            None
            if self.issuer is None
            else (
                CEI.issuer(self.issuer) if isinstance(self.issuer, str) else self.issuer
            )
        )

    def _create_cei_lang_mom(self) -> Optional[etree._Element]:
        return None if self.language is None else CEI.lang_MOM(self.language)

    def _create_cei_list_bibl(self) -> Optional[etree._Element]:
        return (
            CEI.listBibl(*self._create_cei_bibls(self.literature))
            if len(self.literature)
            else None
        )

    def _create_cei_list_bibl_edition(self) -> Optional[etree._Element]:
        return (
            CEI.listBiblEdition(*self._create_cei_bibls(self.literature_editions))
            if len(self.literature_editions)
            else None
        )

    def _create_cei_list_bibl_erw(self) -> Optional[etree._Element]:
        return (
            CEI.listBiblErw(*self._create_cei_bibls(self.literature_secondary))
            if len(self.literature_secondary)
            else None
        )

    def _create_cei_list_bibl_faksimile(self) -> Optional[etree._Element]:
        return (
            CEI.listBiblFaksimile(*self._create_cei_bibls(self.literature_depictions))
            if len(self.literature_depictions)
            else None
        )

    def _create_cei_list_bibl_regest(self) -> Optional[etree._Element]:
        return (
            CEI.listBiblRegest(*self._create_cei_bibls(self.literature_abstracts))
            if len(self.literature_abstracts)
            else None
        )

    def _create_cei_material(self) -> Optional[etree._Element]:
        return None if self.material is None else CEI.material(self.material)

    def _create_cei_nota(self) -> List[etree._Element]:
        return [CEI.nota(nota) for nota in self.chancellary_remarks]

    def _create_cei_notarius_desc(self) -> Optional[etree._Element]:
        return (
            self.notarial_authentication
            if self.notarial_authentication is None
            or isinstance(self.notarial_authentication, etree._Element)
            else CEI.notariusDesc(self.notarial_authentication)
        )

    def _create_cei_geog_name(
        self, value: Optional[str | etree._Element]
    ) -> Optional[etree._Element]:
        return CEI.geogName(value) if isinstance(value, str) else value

    def _create_cei_index(
        self, value: Optional[str | etree._Element]
    ) -> Optional[etree._Element]:
        return CEI.index(value) if isinstance(value, str) else value

    def _create_cei_org_name(
        self, value: Optional[str | etree._Element]
    ) -> Optional[etree._Element]:
        return CEI.orgName(value) if isinstance(value, str) else value

    def _create_cei_p(self) -> List[etree._Element]:
        return (
            [CEI.p(comment) for comment in self.comments] if len(self.comments) else []
        )

    def _create_cei_pers_name(
        self, value: Optional[str | etree._Element], type: Optional[str] = None
    ) -> Optional[etree._Element]:
        if isinstance(value, str):
            attributes = {}
            if type is not None:
                attributes["type"] = type
            return CEI.persName(value, attributes)
        elif isinstance(value, etree._Element):
            if type is not None:
                value.set("type", type)
            return value
        else:
            return None

    def _create_cei_physical_desc(self) -> Optional[etree._Element]:
        children = join(
            self._create_cei_material(),
            self._create_cei_dimensions(),
            self._create_cei_condition(),
        )
        return CEI.physicalDesc(*children) if len(children) else None

    def _create_cei_place_name(
        self, value: Optional[str | etree._Element]
    ) -> Optional[etree._Element]:
        return CEI.placeName(value) if isinstance(value, str) else value

    def _create_cei_quote_originaldatierung(self) -> Optional[etree._Element]:
        return (
            self.date_quote
            if self.date_quote is None or isinstance(self.date_quote, etree._Element)
            else CEI.quoteOriginaldatierung(self.date_quote)
        )

    def _create_cei_recipient(self) -> Optional[etree._Element]:
        return (
            None
            if self.recipient is None
            else (
                CEI.recipient(self.recipient)
                if isinstance(self.recipient, str)
                else self.recipient
            )
        )

    def _create_cei_ref(self) -> Optional[etree._Element]:
        return (
            None
            if self.external_link is None
            else CEI.ref({"target": self.external_link})
        )

    def _create_cei_seal_desc(self) -> Optional[etree._Element]:
        if self.seals is None:
            return None
        elif isinstance(self.seals, etree._Element):
            return self.seals
        elif isinstance(self.seals, str):
            return CEI.sealDesc(self.seals)
        elif isinstance(self.seals, Seal):
            return CEI.sealDesc(self.seals.to_xml())
        else:
            # List of strings or Seal objects
            return CEI.sealDesc(
                *[
                    CEI.seal(desc) if isinstance(desc, str) else desc.to_xml()
                    for desc in self.seals
                ]
            )

    def _create_cei_source_desc(self) -> Optional[etree._Element]:
        children = []
        if self.abstract_sources:
            children.append(
                CEI.sourceDescRegest(*self._create_cei_bibls(self.abstract_sources))
            )
        if self.transcription_sources:
            children.append(
                CEI.sourceDescVolltext(
                    *self._create_cei_bibls(self.transcription_sources)
                )
            )
        return CEI.sourceDesc(*children) if len(children) else None

    def _create_cei_tenor(self) -> Optional[etree._Element]:
        return (
            self.transcription
            if self.transcription is None
            or isinstance(self.transcription, etree._Element)
            else CEI.tenor(self.transcription)
        )

    def _create_cei_text(self, add_schema_location: bool = False) -> etree._Element:
        text = CEI.text(
            self._create_cei_front(),
            self._create_cei_body(),
            self._create_cei_back(),
            type="charter",
        )
        if add_schema_location:
            text.attrib.update(CEI_SCHEMA_LOCATION_ATTRIBUTE)
        return text

    def _create_cei_traditio_form(self) -> Optional[etree._Element]:
        return None if not self._tradition else CEI.traditioForm(self._tradition)

    def _create_cei_witness_orig(self) -> Optional[etree._Element]:
        children = join(
            self._create_cei_traditio_form(),
            self._create_cei_arch_identifier(),
            self._create_cei_auth(),
            self._create_cei_physical_desc(),
            self._create_cei_nota(),
            self._create_cei_figures(),
        )
        return CEI.witnessOrig(*children) if len(children) else None

    # --------------------------------------------------------------------#
    #                           Public methods                           #
    # --------------------------------------------------------------------#

    def to_xml(self, add_schema_location: bool = False) -> etree._Element:
        """Creates an xml representation of the charter.

        Args:
            add_schema_location: If True, the CEI schema location is added to the root element.

        Returns:
            An etree Element object representing the charter.
        """
        return self._create_cei_text(add_schema_location)

    def to_file(self, folder: Optional[str] = None, add_schema_location: bool = False):
        """Writes the xml representation of the charter to a file. The filename is generated from the normalized charter id.

        Args:
            folder (str): The folder to write the file to. If this is ommitted, the file is written to the place where the script is executed from.
            add_schema_location (bool): If True, the CEI schema location is added to the root element. Defaults to False.
        """
        return super(Charter, self).to_file(
            self.id_norm + ".cei",
            folder=folder,
            add_schema_location=add_schema_location,
        )



import pytest

from to_cei.validator import Validator


def test_no_exception_for_correct_cei(valid_cei):
    Validator().validate_cei(valid_cei)


def test_raises_exception_for_incorrect_cei(invalid_cei):
    with pytest.raises(Exception):
        Validator().validate_cei(invalid_cei)



import pathlib

import pytest
from lxml import etree

from to_cei.charter import Charter
from to_cei.charter_group import CharterGroup
from to_cei.config import SCHEMA_LOCATION, SCHEMA_LOCATION_QNAME
from to_cei.validator import Validator


def test_is_valid_cei():
    group = CharterGroup("Charter group", [Charter("1A"), Charter("1b")])
    Validator().validate_cei(group.to_xml())


def test_writes_correct_file(tmp_path):
    d = tmp_path
    group = CharterGroup("Charter group", [Charter("1A"), Charter("1b")])
    group.to_file(d)
    out = pathlib.Path(d, "charter_group.cei.group.xml")
    assert out.is_file()
    written = etree.parse(str(out))
    Validator().validate_cei(written.getroot())


def test_add_schema_location_is_respected():
    group = CharterGroup("Charter group", [Charter("1A"), Charter("1b")])
    assert (
        group.to_xml(add_schema_location=True).get(SCHEMA_LOCATION_QNAME)
        == SCHEMA_LOCATION
    )
    assert group.to_xml(add_schema_location=False).get(SCHEMA_LOCATION_QNAME) == None


def test_raises_exception_for_empty_name():
    with pytest.raises(ValueError):
        CharterGroup("")



import os
import shelve
from pathlib import Path
from typing import Optional

import requests


class FileCache:
    __shelf: shelve.Shelf

    def __init__(self, base: str = str(Path.home().joinpath(".cache", "to-cei"))):
        if not os.path.isdir(base):
            os.makedirs(base)
        file = os.path.join(base, "cache")
        self.__shelf = shelve.open(file, writeback=True)

    def __exit__(self):
        self.__shelf.close()

    def get(self, url: str, force: bool = False) -> Optional[str]:
        if url not in self.__shelf or force == True:
            with requests.get(url) as response:
                self.__shelf[url] = response.text
                return response.text
        else:
            text = self.__shelf[url]
            assert isinstance(text, str)
            return text



from typing import List, Optional

from lxml import etree

from to_cei.config import CEI_NS


def join(
    *values: Optional[etree._Element | List[etree._Element]],
) -> List[etree._Element]:
    """Joins all non-empty values in a list."""
    all = []
    for value in values:
        if isinstance(value, etree._Element):
            all.append(value)
        elif isinstance(value, List) and len(value):
            all = all + value
    return all


def ln(element: etree._Element) -> str:
    """Get the local name of an element."""
    return etree.QName(element.tag).localname


def ns(element: etree._Element) -> str:
    """Get the namespace of an element."""
    return etree.QName(element.tag).namespace


def get_str(value: Optional[str] = None) -> Optional[str]:
    return value if value is not None and len(value) else None


def get_str_list(value: Optional[str | List[str]] = []) -> List[str]:
    return (
        []
        if value is None or (isinstance(value, str) and not len(value))
        else (value if isinstance(value, List) else [value])
    )


def get_str_or_element(
    value: Optional[str | etree._Element], *tags: str
) -> Optional[str | etree._Element]:
    if isinstance(value, str) and not len(value):
        return None
    if isinstance(value, etree._Element):
        if ns(value) != CEI_NS:
            raise ValueError(
                "Provided element needs to be in the CEI namespace but instead is in '{}'".format(
                    ns(value)
                )
            )
        if ln(value) not in tags:
            raise ValueError(
                "Provided element needs to be one of '{}', but instead is '{}'".format(
                    ", ".join(tags), ln(value)
                )
            )
    return value


def get_str_or_element_list(
    values: Optional[List[str | etree._Element]], *tags: str
) -> List[str | etree._Element]:
    result = []
    if values is not None:
        for value in values:
            value = get_str_or_element(value, *tags)
            if value is not None:
                result.append(value)
    return result



from lxml import etree
from lxml.builder import ElementMaker  # type: ignore

from to_cei import filecache

file_cache = filecache.FileCache()

CEI_NS: str = "http://www.monasterium.net/NS/cei"

CEI_PREFIX: str = "cei"

CHARTER_NSS = {
    CEI_PREFIX: CEI_NS,
}

SCHEMA_LOCATION_QNAME = etree.QName(
    "http://www.w3.org/2001/XMLSchema-instance", "schemaLocation"
)

SCHEMA_LOCATION = f"{CEI_NS} {CEI_NS}"

CEI_SCHEMA_LOCATION_ATTRIBUTE = {SCHEMA_LOCATION_QNAME: SCHEMA_LOCATION}

XSI_NS = "http://www.w3.org/2001/XMLSchema-instance"


CEI = ElementMaker(namespace=CEI_NS, nsmap=CHARTER_NSS)



from codecs import open
from os import path

from setuptools import setup

pwd = path.abspath(path.dirname(__file__))

with open(path.join(pwd, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

# This call to setup() does all the work
setup(
    name="to_cei",
    version="0.1.6",
    description="to-CEI",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/icaruseu/to-cei",
    author="Daniel Jeller",
    author_email="it@icar-us.eu",
    license="GPL 3.0",
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "Operating System :: OS Independent",
    ],
    packages=["to_cei"],
    include_package_data=True,
    install_requires=["astropy", "lxml", "requests", "xmlschema"],
)



import os
from abc import ABC, abstractmethod
from pathlib import Path
from typing import List, Optional

from lxml import etree

from to_cei.config import CEI_PREFIX


class XmlAssembler(ABC):
    @abstractmethod
    def to_xml(self, add_schema_location: bool = False) -> Optional[etree._Element]:
        pass

    def to_string(self, add_schema_location: bool = False) -> str:
        """Serializes the xml representation of the object to a string.

        Args:
            add_schema_location: If True, the CEI schema location is added to the root element.

        Returns:
            A string representation of the object.
        """
        xml = self.to_xml(add_schema_location)
        return (
            ""
            if xml is None
            else str(
                etree.tostring(
                    xml,
                    encoding="unicode",
                    pretty_print=True,
                )
            )
        )

    def to_file(
        self,
        name: str,
        folder: Optional[str | Path] = None,
        inclusive_ns_prefixes: List[str] = [],
        add_schema_location: bool = False,
    ):
        xml = self.to_xml(add_schema_location)
        if xml is None:
            raise Exception("Failed to read xml")
        folder = (
            folder
            if isinstance(folder, str)
            else (
                str(folder.absolute())
                if isinstance(folder, Path)
                else os.path.abspath(os.getcwd())
            )
        )
        os.makedirs(folder, exist_ok=True)
        etree.ElementTree(xml).write(
            os.path.join(folder, name + ".xml"),
            encoding="UTF-8",
            pretty_print=True,
            inclusive_ns_prefixes=[CEI_PREFIX] + inclusive_ns_prefixes,
            xml_declaration=True,
            standalone=False,
        )



import pytest

from pytest_helpers import xp, xps
from to_cei.config import CEI
from to_cei.seal import Seal
from to_cei.validator import Validator


def test_is_seal_valid():
    seal = Seal(
        condition="A condotion",
        dimensions="Some dimensions",
        legend=[("recto", "The recto legend"), ("verso", "The verso legend")],
        material="A material",
        sigillant="The sigillant",
    )
    Validator().validate_cei(
        CEI.text(
            CEI.front(),
            CEI.body(
                CEI.idno("1", {"id": "1"}),
                CEI.chDesc(CEI.witnessOrig(CEI.auth(CEI.sealDesc(seal.to_xml())))),
            ),
            CEI.back(),
            {"type": "charter"},
        )
    )


def test_correctly_returns_no_xml_for_empty_seal():
    assert Seal().to_xml() is None


def test_has_correct_condition():
    condition = "A condition"
    seal = Seal(
        condition=condition,
    )
    assert seal.condition == condition
    condition_xml = xps(seal, "/cei:seal/cei:sealCondition")
    assert condition_xml.text == condition


def test_has_correct_dimensions():
    dimensions = "A dimension"
    seal = Seal(
        dimensions=dimensions,
    )
    assert seal.dimensions == dimensions
    dimensions_xml = xps(seal, "/cei:seal/cei:sealDimensions")
    assert dimensions_xml.text == dimensions


def test_has_correct_single_legend():
    legend = "A single legend"
    seal = Seal(
        legend=legend,
    )
    assert seal.legend == legend
    legend_xml = xps(seal, "/cei:seal/cei:legend")
    assert legend_xml.text == legend


def test_has_correct_multiple_legends():
    legends = [("recto", "The recto legend"), ("verso", "The verso legend")]
    seal = Seal(
        legend=legends,
    )
    assert seal.legend == legends
    legends_xml = xp(seal, "/cei:seal/cei:legend")
    assert len(legends_xml) == 2
    assert legends_xml[0].text == legends[0][1]
    assert legends_xml[0].get("place") == legends[0][0]
    assert legends_xml[1].text == legends[1][1]
    assert legends_xml[1].get("place") == legends[1][0]


def test_has_correct_material():
    material = "A material"
    seal = Seal(
        material=material,
    )
    assert seal.material == material
    material_xml = xps(seal, "/cei:seal/cei:sealMaterial")
    assert material_xml.text == material


def test_has_correct_text_sigillant():
    sigillant = "A sigillant"
    seal = Seal(
        sigillant=sigillant,
    )
    assert seal.sigillant == sigillant
    sigillant_xml = xps(seal, "/cei:seal/cei:sigillant")
    assert sigillant_xml.text == sigillant


def test_has_correct_xml_sigillant():
    sigillant = CEI.persName("A sigillant")
    seal = Seal(sigillant=sigillant)
    assert seal.sigillant == sigillant
    sigillant_xml = xps(seal, "/cei:seal/cei:sigillant/cei:persName")
    assert sigillant_xml == sigillant


def test_raises_exception_for_wrong_sigillant_xml():
    with pytest.raises(ValueError):
        Seal(sigillant=CEI.placeName())



from lxml import etree

from to_cei.config import CEI, CEI_NS
from to_cei.helpers import join, ln, ns


def test_gets_correct_local_name():
    assert ln(CEI.text()) == "text"


def test_gets_correct_namespace():
    assert ns(CEI.text()) == CEI_NS


def test_joins_correctly():
    joined = join(
        CEI.text(), None, CEI.persName(), None, [], [CEI.placeName(), CEI.persName()]
    )
    assert len(joined) == 4
    assert etree.tostring(joined[0]) == etree.tostring(CEI.text())
    assert etree.tostring(joined[1]) == etree.tostring(CEI.persName())
    assert etree.tostring(joined[2]) == etree.tostring(CEI.placeName())
    assert etree.tostring(joined[3]) == etree.tostring(CEI.persName())
